import numpy as np
import SimpleITK as sitk
import CircuitSeeker.utility as ut
import os, psutil
from scipy.ndimage import map_coordinates


def apply_transform(
    fix, mov,
    fix_spacing, mov_spacing,
    transform_list,
    transform_spacing=None,
    fix_origin=None,
    mov_origin=None,
    interpolate_with_nn=False,
    extrapolate_with_nn=False,
):
    """
    """

    # set global number of threads
    if "LSB_DJOB_NUMPROC" in os.environ:
        ncores = int(os.environ["LSB_DJOB_NUMPROC"])
    else:
        ncores = psutil.cpu_count(logical=False)
    sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(2*ncores)

    # convert images to sitk objects
    dtype = fix.dtype
    fix = ut.numpy_to_sitk(fix, fix_spacing, fix_origin)
    mov = ut.numpy_to_sitk(mov, mov_spacing, mov_origin)

    # construct transform
    transform = sitk.CompositeTransform(3)
    for i, t in enumerate(transform_list):

        # affine transforms
        if len(t.shape) == 2:
            t = ut.matrix_to_affine_transform(t)

        # bspline parameters
        elif len(t.shape) == 1:
            t = ut.bspline_parameters_to_transform(t)

        # fields
        elif len(t.shape) == 4:
            # set transform_spacing
            if transform_spacing is None:
                sp = fix_spacing
            elif isinstance(transform_spacing[i], tuple):
                sp = transform_spacing[i]
            else:
                sp = transform_spacing
            # create field
            t = ut.field_to_displacement_field_transform(t, sp)

        # add to composite transform
        transform.AddTransform(t)

    # set up resampler object
    resampler = sitk.ResampleImageFilter()
    resampler.SetNumberOfThreads(2*ncores)
    resampler.SetReferenceImage(sitk.Cast(fix, sitk.sitkFloat32))
    resampler.SetTransform(transform)

    # check for NN interpolation
    if interpolate_with_nn:
        resampler.SetInterpolator(sitk.sitkNearestNeighbor)

    # check for NN extrapolation
    if extrapolate_with_nn:
        resampler.SetUseNearestNeighborExtrapolator(True)

    # execute, return as numpy array
    resampled = resampler.Execute(sitk.Cast(mov, sitk.sitkFloat32))
    return sitk.GetArrayFromImage(resampled).astype(dtype)


def apply_transform_to_coordinates(
    coordinates,
    transform,
    transform_spacing=None,
):
    """
    """

    # if transform is an affine matrix
    if transform.shape == (4, 4):

        # matrix vector multiply
        mm, tt = transform[:3, :3], transform[:3, -1]
        return np.einsum('...ij,...j->...i', mm, coordinates) + tt

    # if transform is a deformation vector field
    else:

        # transform_spacing must be given
        error_message = "If transform is a displacement vector field, "
        error_message += "transform_spacing must be given."
        assert (transform_spacing is not None), error_message

        # get coordinates in transform voxel units, reformat for map_coordinates
        coordinates = ( coordinates / transform_spacing ).transpose()

        # interpolate position field at coordinates, reformat, return
        interp = lambda x: map_coordinates(x, coordinates, order=1, mode='nearest')
        dX = np.array([interp(transform[..., i]) for i in range(3)]).transpose()
        return coordinates.transpose() * transform_spacing + dX


# TODO: convert this and following function into single function "compose"
#       that checks types with conditionals and does the right thing
def compose_displacement_vector_fields(
    first_field,
    second_field,
    spacing,
):
    """
    """

    # container for warped first field
    first_field_warped = np.empty_like(first_field)

    # loop over components
    for iii in range(3):

        # warp first field with second
        first_field_warped[..., iii] = apply_transform(
            first_field[..., iii], first_field[..., iii],
            spacing, spacing,
            transform_list=[second_field,],
            extrapolate_with_nn=True,
        )

    # combine warped first field and second field
    return first_field_warped + second_field


def compose_affine_and_displacement_vector_field(
    transform_one,
    transform_two,
    spacing,
):
    """
    """

    # reformat affine as displacement vector field
    if transform_one.shape == (4, 4):
        transform_one = ut.matrix_to_displacement_field(
            transform_two[..., 0], transform_one, spacing,
        )
    else:
        transform_two = ut.matrix_to_displacement_field(
            transform_one[..., 0], transform_two, spacing,
        )

    # compose fields, return
    return compose_displacement_vector_fields(
        transform_one, transform_two, spacing,
    )


def invert_displacement_vector_field(
    field,
    spacing,
    iterations=10,
    order=2,
    sqrt_iterations=5,
):
    """
    """

    # initialize inverse as negative root
    root = _displacement_field_composition_nth_square_root(
        field, spacing, order, sqrt_iterations,
    )
    inv = - np.copy(root)

    # iterate to invert
    for i in range(iterations):
        inv -= compose_displacement_vector_fields(root, inv, spacing)

    # square-compose inv order times
    for i in range(order):
        inv = compose_displacement_vector_fields(inv, inv, spacing)

    # return result
    return inv


def _displacement_field_composition_nth_square_root(
    field,
    spacing,
    order,
    sqrt_iterations=5,
):
    """
    """

    # initialize with given field
    root = np.copy(field)

    # iterate taking square roots
    for i in range(order):
        root = _displacement_field_composition_square_root(
            root, spacing, iterations=sqrt_iterations,
        )

    # return result
    return root


def _displacement_field_composition_square_root(
    field,
    spacing,
    iterations=5,
):
    """
    """

    # container to hold root
    root = 0.5 * field

    # iterate
    for i in range(iterations):
        residual = (field - compose_displacement_vector_fields(root, root, spacing))
        root += 0.5 * residual

    # return result
    return root





import os, psutil
import SimpleITK as sitk


def configure_irm(
    metric='MI',
    bins=128,
    sampling='regular',
    sampling_percentage=1.0,
    optimizer='GD',
    iterations=200,
    learning_rate=1.0,
    estimate_learning_rate="once",
    min_step=0.1,
    max_step=1.0,
    shrink_factors=[2,1],
    smooth_sigmas=[2.,1.],
    num_steps=[2, 2, 2],
    step_sizes=[1., 1., 1.],
    callback=None,
):
    """
    Wrapper exposing some of the itk::simple::ImageRegistrationMethod API
    Rarely called by the user. Typically used in custom registration functions.

    Parameters
    ----------
    metric : string (default: 'MI')
        The image matching term optimized during alignment
        Options:
            'MI': mutual information
            'CC': correlation coefficient
            'MS': mean squares

    bins : int (default: 128)
        Only used when `metric`='MI'. Number of histogram bins
        for image intensity histograms. Ignored when `metric` is
        'CC' or 'MS'

    sampling : string (default: 'regular')
        How image intensities are sampled during metric calculation
        Options:
            'regular': sample intensities with regular spacing
            'random': sample intensities randomly

    sampling_percentage : float in range [0., 1.] (default: 1.0)
        Percentage of voxels used during metric sampling

    optimizer : string (default 'GD')
        Optimization algorithm used to find a transform
        Options:
            'GD': gradient descent
            'RGD': regular gradient descent
            'EX': exhaustive - regular sampling of transform parameters between
                  given limits

    iterations : int (default: 200)
        Maximum number of iterations at each scale level to run optimization.
        Optimization may still converge early.

    learning_rate : float (default: 1.0)
        Initial gradient descent step size

    estimate_learning_rate : string (default: "once")
        Frequency of estimating the learning rate. Only used if `optimizer`='GD'
        Options:
            'once': only estimate once at the beginning of optimization
            'each_iteration': estimate step size at every iteration
            'never': never estimate step size, `learning_rate` is fixed

    min_step : float (default: 0.1)
        Minimum allowable gradient descent step size. Only used if `optimizer`='RGD'

    max_step : float (default: 1.0)
        Maximum allowable gradient descent step size. Used by both 'GD' and 'RGD'

    shrink_factors : iterable of type int (default: [2, 1])
        Downsampling scale levels at which to optimize

    smooth_sigmas : iterable of type float (default: [2., 1.])
        Sigma of Gaussian used to smooth each scale level image
        Must be same length as `shrink_factors`
        Should be specified in physical units, e.g. mm or um

    num_steps : iterable of type int (default: [2, 2, 2])
        Only used if `optimizer`='EX'
        Number of steps to search in each direction from the initial
        position of the transform parameters

    step_sizes : iterable of type float (default: [1., 1., 1.])
        Only used if `optimizer`='EX'
        Size of step to take during brute force optimization
        Order of parameters and relevant scales should be based on
        the type of transform being optimized

    callable : callable object, e.g. function (default: None)
        A function run at every iteration of optimization
        Should take only the ImageRegistrationMethod object as input: `irm`
        If None then the Level, Iteration, and Metric values are
        printed at each iteration

    Returns
    -------
    irm : itk::simple::ImageRegistrationMethod object
        The configured ImageRegistrationMethod object. Simply needs
        images and a transform type to be ready for optimization.
    """

    # identify number of cores available, assume hyperthreading
    if "LSB_DJOB_NUMPROC" in os.environ:
        ncores = int(os.environ["LSB_DJOB_NUMPROC"])
    else:
        ncores = psutil.cpu_count(logical=False)

    # initialize IRM object, be completely sure nthreads is set
    sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(2*ncores)
    irm = sitk.ImageRegistrationMethod()
    irm.SetNumberOfThreads(2*ncores)

    # set interpolator
    irm.SetInterpolator(sitk.sitkLinear)

    # set metric
    if metric == 'MI':
        irm.SetMetricAsMattesMutualInformation(
            numberOfHistogramBins=bins,
        )
    elif metric == 'CC':
        irm.SetMetricAsCorrelation()
    elif metric == 'MS':
        irm.SetMetricAsMeanSquares()

    # set metric sampling type and percentage
    if sampling == 'regular':
        irm.SetMetricSamplingStrategy(irm.REGULAR)
    elif sampling == 'random':
        irm.SetMetricSamplingStrategy(irm.RANDOM)
    irm.SetMetricSamplingPercentage(sampling_percentage)

    # set estimate learning rate
    if estimate_learning_rate == "never":
        estimate_learning_rate = irm.Never
    elif estimate_learning_rate == "once":
        estimate_learning_rate = irm.Once
    elif estimate_learning_rate == "each_iteration":
        estimate_learning_rate = irm.EachIteration

    # set optimizer
    if optimizer == 'GD':
        irm.SetOptimizerAsGradientDescent(
            numberOfIterations=iterations,
            learningRate=learning_rate,
            maximumStepSizeInPhysicalUnits=max_step,
            estimateLearningRate=estimate_learning_rate,
        )
        irm.SetOptimizerScalesFromPhysicalShift()
    elif optimizer == 'RGD':
        irm.SetOptimizerAsRegularStepGradientDescent(
            minStep=min_step, learningRate=learning_rate,
            numberOfIterations=iterations,
            maximumStepSizeInPhysicalUnits=max_step,
        )
        irm.SetOptimizerScalesFromPhysicalShift()
    elif optimizer == 'EX':
        irm.SetOptimizerAsExhaustive(num_steps[::-1])
        irm.SetOptimizerScales(step_sizes[::-1])

    # set pyramid
    irm.SetShrinkFactorsPerLevel(shrinkFactors=shrink_factors)
    irm.SetSmoothingSigmasPerLevel(smoothingSigmas=smooth_sigmas)
    irm.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

    # set callback function
    if callback is None:
        def callback(irm):
            level = irm.GetCurrentLevel()
            iteration = irm.GetOptimizerIteration()
            metric = irm.GetMetricValue()
            print("LEVEL: ", level, " ITERATION: ", iteration, " METRIC: ", metric)
    irm.AddCommand(sitk.sitkIterationEvent, lambda: callback(irm))

    # return configured irm
    return irm




import numpy as np
from scipy.ndimage import zoom
from scipy.ndimage.filters import gaussian_filter
from scipy.ndimage.morphology import binary_erosion, binary_dilation, binary_fill_holes
from scipy.ndimage.measurements import label, labeled_comprehension
import morphsnakes


def estimate_background(image, rad=5):
    """
    """

    a, b = slice(0, rad), slice(-rad, None)
    corners = ((a,a,a), (a,a,b), (a,b,a), (a,b,b),
               (b,a,a), (b,a,b), (b,b,a), (b,b,b))
    return np.median([np.mean(image[c]) for c in corners])


def segment(
    image,
    lambda2,
    iterations,
    smoothing=1,
    threshold=None,
    init=None
    ):
    """
    """

    if threshold is not None:
        image[image < threshold] = 0
    if init is None:
        init = np.zeros_like(image, dtype=np.uint8)
        bounds = np.ceil(np.array(image.shape) * 0.1).astype(int)
        init[tuple(slice(b, -b) for b in bounds)] = 1
    return morphsnakes.morphological_chan_vese(
        image,
        iterations,
        init_level_set=init,
        smoothing=smoothing,
        lambda2=lambda2,
    ).astype(np.uint8)


def largest_connected_component(mask):
    lbls, nlbls = label(mask)
    vols = labeled_comprehension(mask, lbls, range(1, nlbls+1), np.sum, float, 0)
    mask[lbls != np.argmax(vols)+1] = 0
    return mask


def brain_detection(
    image,
    voxel_spacing,
    iterations=[40,8,2],
    shrink_factors=[4,2,1],
    smooth_sigmas=[8,4,2],
    lambda2=20,
    background=None,
    mask=None,
    mask_smoothing=1,
    ):
    """
    """

    # segment
    if background is None:
        background = estimate_background(image)
    for its, sf, ss in zip(iterations, shrink_factors, smooth_sigmas):
        image_small = zoom(gaussian_filter(image, ss/voxel_spacing), 1./sf, order=1)
        if mask is not None:
            zoom_factors = [x/y for x, y in zip(image_small.shape, mask.shape)]
            mask = zoom(mask, zoom_factors, order=0)
        mask = segment(
            image_small,
            lambda2=lambda2,
            iterations=its,
            smoothing=mask_smoothing,
            threshold=background,
            init=mask,
        )

    # basic topological correction
    mask = binary_erosion(mask, iterations=2)
    mask = largest_connected_component(mask)
    mask = binary_dilation(mask, iterations=2)
    mask = binary_fill_holes(mask).astype(np.uint8)

    # ensure output is on correct grid
    if mask.shape != image.shape:
        zoom_factors = [x/y for x, y in zip(image.shape, mask.shape)]
        mask = zoom(mask, zoom_factors, order=0)
    return mask





import numpy as np
import zarr
import ClusterWrap
import dask.array as da
from numcodecs import Blosc


def deltafoverf(
    array,
    window_size,
):
    """
    """

    # create container to store result
    new_shape = (array.shape[0] - window_size,) + array.shape[1:]
    dff = np.empty(new_shape, dtype=np.float16)  # TODO: user specifies precision

    # get window start and stop indices and sum image
    w_start, w_stop = 0, window_size
    sum_image = np.sum(array[w_start:w_stop], axis=0)

    # loop over frames
    for iii in range(window_size, array.shape[0]):
        baseline = sum_image / window_size
        dff[iii - window_size] = (array[iii] - baseline) / (baseline + 1)
        sum_image = sum_image - array[w_start] + array[iii]
        w_start += 1

    # return result
    return dff


def distributed_deltafoverf(
    zarr_path,
    window_size,
    batch_size,
    write_path,
    compression_level=4,
    cluster_kwargs={},
):
    """
    """

    # launch cluster
    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # lazy load zarr to get metadata
        metadata = zarr.open(zarr_path, 'r')

        # get block start indices
        start_indices, start_index = [], 0
        while start_index + window_size < metadata.shape[0]:
            start_indices.append(start_index)
            start_index = start_index + batch_size - window_size

        # convert to dask array
        start_indices_da = da.from_array(start_indices, chunks=(1,))

        # wrap deltafoverf function
        def wrapped_deltafoverf(index):
            zarr_file = zarr.open(zarr_path, 'r')
            data = zarr_file[index[0]:index[0]+batch_size]
            return deltafoverf(data, window_size)
 
        # map function to each block
        dff = da.map_blocks(
            wrapped_deltafoverf, start_indices_da,
            dtype=np.float16,
            new_axis=list(range(1, metadata.ndim)),
            chunks=(batch_size-window_size,) + metadata.chunks[1:],
        )

        # ensure the correct shape and rechunk for faster writing
        dff = dff[:metadata.shape[0] - window_size]
        dff = dff.rechunk((1,) + metadata.chunks[1:])

        # persist dff before writing to zarr, prevents RAM conflicts
        dff = dff.persist()

        # write to output zarr
        compressor = Blosc(
            cname='zstd',
            clevel=compression_level,
            shuffle=Blosc.BITSHUFFLE,
        )
        dff_disk = zarr.open(
            write_path, 'w',
            shape=dff.shape,
            chunks=metadata.chunks,
            dtype=dff.dtype,
            compressor=compressor,
        )
        da.to_zarr(dff, dff_disk)

        # return reference to zarr store
        return dff_disk




import numpy as np
import os
import ClusterWrap
import CircuitSeeker.fileio as csio
import CircuitSeeker.utility as ut
from CircuitSeeker.align import affine_align
from CircuitSeeker.align import deformable_align
from CircuitSeeker.transform import apply_transform
import dask.array as da
import dask.delayed as delayed
from scipy.ndimage import median_filter
from scipy.ndimage import gaussian_filter1d
from scipy.ndimage import map_coordinates
from scipy.ndimage import zoom
import zarr
from numcodecs import Blosc
from glob import glob
import json


def distributed_image_mean(
    frames,
    cluster_kwargs={},
):
    """
    Voxelwise mean over all images specified by frames.
    Distributed with Dask for expediency.

    Parameters
    ----------
    frames : dict
        Specifies the image set on disk to be averaged. At least three
        keys must be defined:
            'folder' : directory containing the image set
            'prefix' : common prefix to all images being averaged
            'suffix' : common suffix - typically the file extension
        Common values for suffix are '.h5', '.stack', or '.tiff'

        If suffix is '.h5' then an additional key must be defined:
            'dataset_path' : path to image dataset within hdf5 container

        If suffix is '.stack' then two additional keys must be defined:
            'dtype' : a numpy datatype object for the datatype in the raw images
            'shape' : the array shape of the data as a tuple

    cluster_kwargs : dict (default: {})
        Arguments passed to ClusterWrap.cluster
        If working with an LSF cluster, this will be
        ClusterWrap.janelia_lsf_cluster. If on a workstation
        this will be ClusterWrap.local_cluster.
        This is how distribution parameters are specified.
    """

    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # hdf5 files use dask.array
        if csio.testPathExtensionForHDF5(frames['suffix']):
            frames = csio.daskArrayBackedByHDF5(
                frames['folder'], frames['prefix'],
                frames['suffix'], frames['dataset_path'],
            )
            frames_mean = frames.mean(axis=0, dtype=np.float32).compute()
            frames_mean = np.round(frames_mean).astype(frames[0].dtype)
        # stack files use dask.array
        elif csio.testPathExtensionForSTACK(frames['suffix']):
            frames = csio.daskArrayBackedBySTACK(
                frames['folder'], frames['prefix'], frames['suffix'],
                frames['dtype'], frames['shape'],
            )
            frames_mean = frames.mean(axis=0, dtype=np.float32).compute()
            frames_mean = np.round(frames_mean).astype(frames[0].dtype)
        # other types use dask.bag
        else:
            frames = csio.daskBagOfFilePaths(
                frames['folder'], frames['prefix'], frames['suffix'],
            )
            nframes = frames.npartitions
            frames_mean = frames.map(csio.readImage).reduction(sum, sum).compute()
            dtype = frames_mean.dtype
            frames_mean = np.round(frames_mean/np.float(nframes)).astype(dtype)

    # return reference to mean image
    return frames_mean


def motion_correct(
    fix, frames,
    fix_spacing, frames_spacing,
    time_stride=1,
    sigma=7,
    fix_mask=None,
    cluster_kwargs={},
    **kwargs,
):
    """
    """

    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # scatter fixed image to cluster
        fix_d = cluster.client.scatter(fix, broadcast=True)
        # scatter fixed mask to cluster
        fix_mask_d = None
        if fix_mask is not None:
            fix_mask_d = cluster.client.scatter(fix_mask, broadcast=True)

        # get total number of frames
        total_frames = len(csio.globPaths(
            frames['folder'], frames['prefix'], frames['suffix'],
        ))

        # crate dask array of all frames
        if csio.testPathExtensionForHDF5(frames['suffix']):
            frames_data = csio.daskArrayBackedByHDF5(
                frames['folder'], frames['prefix'],
                frames['suffix'], frames['dataset_path'],
                stride=time_stride,
            )
        elif csio.testPathExtensionForSTACK(frames['suffix']):
            frames_data = csio.daskArrayBackedBySTACK(
                frames['folder'], frames['prefix'], frames['suffix'],
                frames['dtype'], frames['shape'],
                stride=time_stride,
            )
        compute_frames = frames_data.shape[0]

        # set alignment defaults
        alignment_defaults = {
            'rigid':True,
            'alignment_spacing':2.0,
            'metric':'MS',
            'sampling':'random',
            'sampling_percentage':0.1,
            'optimizer':'RGD',
            'iterations':50,
            'max_step':2.0,
        }
        for k, v in alignment_defaults.items():
            if k not in kwargs:
                kwargs[k] = v
 
        # wrap align function
        def wrapped_affine_align(mov, fix_d, fix_mask_d):
            mov = mov.squeeze()
            t = affine_align(
                fix_d, mov, fix_spacing, frames_spacing,
                fix_mask=fix_mask_d,
                **kwargs,
            )
            e = ut.matrix_to_euler_transform(t)
            p = ut.euler_transform_to_parameters(e)
            return p[None, :]

        params = da.map_blocks(
            wrapped_affine_align, frames_data,
            fix_d=fix_d,
            fix_mask_d=fix_mask_d,
            dtype=np.float64,
            drop_axis=[2, 3,],
            chunks=[1, 6],
        ).compute()

    # (weak) outlier removal and smoothing
    params = median_filter(params, footprint=np.ones((3,1)))
    params = gaussian_filter1d(params, sigma, axis=0)

    # interpolate
    if time_stride > 1:
        x = np.linspace(0, compute_frames-1, total_frames)
        coords = np.meshgrid(x, np.mgrid[:6], indexing='ij')
        params = map_coordinates(params, coords)

    # convert to matrices
    transforms = np.empty((total_frames, 4, 4))
    for i in range(params.shape[0]):
        e = ut.parameters_to_euler_transform(params[i])
        t = ut.affine_transform_to_matrix(e)
        transforms[i] = t

    # return all transforms
    return transforms


def deformable_motion_correct(
    fix, frames,
    fix_spacing, frames_spacing,
    time_stride=1,
    sigma=7,
    fix_mask=None,
    affine_kwargs={},
    bspline_kwargs={},
    cluster_kwargs={},
):
    """
    """

    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # wrap fixed data as delayed object
        fix_d = delayed(fix)

        # wrap fixx mask if given
        fix_mask_d = delayed(np.ones(fix.shape, dtype=np.uint8))
        if fix_mask is not None:
            fix_mask_d = delayed(fix_mask)

        # get total number of frames
        total_frames = len(csio.globPaths(
            frames['folder'], frames['prefix'], frames['suffix'],
        ))

        # create dask array of all frames
        frames_data = csio.daskArrayBackedByHDF5(
            frames['folder'], frames['prefix'],
            frames['suffix'], frames['dataset_path'],
            stride=time_stride,
        )
        compute_frames = frames_data.shape[0]

        # affine defaults
        affine_defaults = {
            'alignment_spacing':2.0,
            'metric':'MS',
            'sampling':'random',
            'sampling_percentage':0.1,
            'optimizer':'RGD',
            'iterations':50,
            'max_step':2.0,
        }
        for k, v in affine_defaults.items():
            if k not in affine_kwargs: affine_kwargs[k] = v

        # bspline defaults
        bspline_defaults = {
            'alignment_spacing':1.0,
            'metric':'MI',
            'sampling':'random',
            'sampling_percentage':0.01,
            'iterations':250,
            'shrink_factors':[2,],
            'smooth_sigmas':[2,],
            'max_step':1.0,
            'control_point_spacing':100.,
            'control_point_levels':[1,],
        }
        for k, v in bspline_defaults.items():
            if k not in bspline_kwargs: bspline_kwargs[k] = v
 
        # wrap align function
        def wrapped_bspline_align(mov, fix_d, fix_mask_d):
            mov = mov.squeeze()
            a = affine_align(
                fix_d, mov, fix_spacing, frames_spacing,
                **affine_kwargs,
            )
            b = bspline_deformable_align(
                fix_d, mov, fix_spacing, frames_spacing,
                fix_mask=fix_mask_d,
                initial_transform=a,
                return_parameters=True,
                **bspline_kwargs,
            )
            return np.hstack((a.flatten(), b))[None, :]

        # total number of params
        # 16 for affine, 18 for bspline fixed params
        # need to compute number of bspline control point params
        xxx = fix.shape * fix_spacing
        y = bspline_kwargs['control_point_spacing']
        cp_grid = [max(1, int(x/y)) + 3 for x in xxx]
        n_params = 16 + 18 + np.prod(cp_grid)*3

        # execute
        params = da.map_blocks(
            wrapped_bspline_align, frames_data,
            fix_d=fix_d,
            fix_mask_d=fix_mask_d,
            dtype=np.float64,
            drop_axis=[2, 3,],
            chunks=[1, n_params],
        ).compute()

    # (weak) outlier removal and smoothing
    params = median_filter(params, footprint=np.ones((3,1)))
    params = gaussian_filter1d(params, sigma, axis=0)

    # interpolate
    if time_stride > 1:
        x = np.linspace(0, compute_frames-1, total_frames)
        coords = np.meshgrid(x, np.mgrid[:n_params], indexing='ij')
        params = map_coordinates(params, coords, order=1)

    # return all parameters
    return params


def save_transforms(path, transforms):
    """
    """

    n = transforms.shape[0]
    d = {i:transforms[i].tolist() for i in range(n)}
    with open(path, 'w') as f:
        json.dump(d, f, indent=4)


def read_transforms(path):
    """
    """

    with open(path, 'r') as f:
        d = json.load(f)
    return np.array([d[str(i)] for i in range(len(d))])


def resample_frames(
    frames,
    frames_spacing,
    transforms,
    write_path,
    mask=None,
    time_stride=1,
    compression_level=4,
    cluster_kwargs={},
):
    """
    """

    with ClusterWrap.cluster(**cluster_kwargs) as cluster:

        # create dask array of all frames
        if csio.testPathExtensionForHDF5(frames['suffix']):
            frames_data = csio.daskArrayBackedByHDF5(
                frames['folder'], frames['prefix'],
                frames['suffix'], frames['dataset_path'],
                stride=time_stride,
            )
        elif csio.testPathExtensionForSTACK(frames['suffix']):
            frames_data = csio.daskArrayBackedBySTACK(
                frames['folder'], frames['prefix'], frames['suffix'],
                frames['dtype'], frames['shape'],
                stride=time_stride,
            )
        compute_frames = frames_data.shape[0]

        # wrap transforms as dask array
        # extra dimension to match frames_data ndims
        if len(transforms.shape) == 3:
            transforms = transforms[::time_stride, None, :, :]
        elif len(transforms.shape) == 2:
            transforms = transforms[::time_stride, None, None, :]
        transforms_d = da.from_array(transforms, chunks=(1,)+transforms[0].shape)

        # wrap mask
        mask_d = None
        if mask is not None:
            mask_sh, frame_sh = mask.shape, frames_data.shape[1:]
            if mask_sh != frame_sh:
                mask = zoom(mask, np.array(frame_sh) / mask_sh, order=0)
            mask_d = cluster.client.scatter(mask, broadcast=True)

        # wrap transform function
        def wrapped_apply_transform(mov, t, mask_d=None):
            mov = mov.squeeze()
            t = t.squeeze()

            # just an affine matrix
            transform_list = [t,]

            # affine plus bspline
            if len(t.shape) == 1:
                transform_list = [t[:16].reshape((4,4)), t[16:]]

            # apply transform(s)
            aligned = apply_transform(
                mov, mov, frames_spacing, frames_spacing,
                transform_list=transform_list,
            )
            if mask_d is not None:
                aligned = aligned * mask_d
            return aligned[None, ...]

        # apply transform to all frames
        frames_aligned = da.map_blocks(
            wrapped_apply_transform, frames_data, transforms_d,
            mask_d=mask_d,
            dtype=np.uint16,
            chunks=[1,] + list(frames_data.shape[1:]),
        )

        # write in parallel as 4D array to zarr file
        compressor = Blosc(
            cname='zstd',
            clevel=compression_level,
            shuffle=Blosc.BITSHUFFLE,
        )
        aligned_disk = zarr.open(
            write_path, 'w',
            shape=frames_aligned.shape,
            chunks=[1,] + list(frames_data.shape[1:]),
            dtype=frames_aligned.dtype,
            compressor=compressor
        )
        da.to_zarr(frames_aligned, aligned_disk)

        # return reference to zarr store
        return aligned_disk





import os, psutil
import numpy as np
import pyfftw
from itertools import product
from CircuitSeeker.transform import apply_transform


def bounded_fourier_shell_correlation(
    fix,
    mov,
    fix_spacing,
    mov_spacing,
    bounds,
):
    """
    """

    # require same data type
    if fix.dtype != mov.dtype:
        error = "Image must have the same data type"
        raise TypeError(error)

    # ensure datasets are on the same grid and spacing
    if fix.shape != mov.shape or not np.all(fix_spacing == mov_spacing):
        mov = apply_transform(
            fix, mov, fix_spacing, mov_spacing, transform_list=[np.eye(4),],
        )

    # establish data types
    if fix.dtype == np.uint16:
        fix = fix.astype(np.float32)
    if fix.dtype == np.float32:
        cdtype = np.complex64
    elif fix.dtype == np.float64:
        cdtype = np.complex128
    else:
        error = "Images must be uint16, float32, or float64 \n"
        error += "fix dtype is: " + str(fix.dtype) + " mov dtype is: " + str(mov.dtype)
        raise TypeError(error)

    # ensure datatypes stay the same
    if fix.dtype != mov.dtype:
        mov = mov.astype(fix.dtype)

    # determine cores available to set fft threads
    if "LSB_DJOB_NUMPROC" in os.environ:
        ncores = int(os.environ["LSB_DJOB_NUMPROC"])
    else:
        ncores = psutil.cpu_count(logical=False)

    # construct fft object
    out_sh = fix.shape[:-1] + (fix.shape[-1]//2 + 1,)  # assuming real valued input
    inp = pyfftw.empty_aligned(fix.shape, dtype=fix.dtype)
    outp = pyfftw.empty_aligned(out_sh, dtype=cdtype)
    fft = pyfftw.FFTW(inp, outp, axes=list(range(len(fix.shape))), threads=2*ncores)
    
    # we'll need to ensure memory layout is compatible with fft object
    flags_to_set = [f for f in ['C', 'F', 'O', 'W', 'A'] if inp.flags[f]]

    # get image frequencies
    inp[:] = np.require(fix, requirements=flags_to_set)
    fix_fourier = np.copy(fft())
    inp[:] = np.require(mov, requirements=flags_to_set)
    mov_fourier = fft()  # no need to copy the last one

    # get frequency position field amplitudes
    a, b = fix.shape, fix_spacing
    freqs = np.meshgrid(*[np.fft.fftfreq(x, y) for x, y in zip(a, b)], indexing='ij')
    freqs = np.array([f[..., :outp.shape[-1]] for f in freqs])
    freqs = np.linalg.norm(freqs, axis=0)**2

    # return frequency shell correlation
    shell = (freqs >= 1./bounds[1]) * (freqs <= 1./bounds[0])
    a = fix_fourier[shell].flatten()
    b = mov_fourier[shell].flatten()
    return np.real(np.corrcoef(a, b)[0, 1])


def cell_quality_score(
    image,
    spacing,
    bounds,
    shift_radius=1,
):
    """
    """

    # closure for correlation function
    CF = lambda a, b: bounded_fourier_shell_correlation(a, b, spacing, spacing, bounds)

    # define all shifts
    # TODO: does not include secondary diagonals (e.g. in 2d: i+1, j-1)
    shifts = product(range(0, shift_radius+1), repeat=3)

    # get correlation for all shifts
    correlations = []
    for iii, shift in enumerate(shifts):
        if shift == (0, 0, 0): continue
        slice_a = tuple(slice(0, -s) if s != 0 else slice(0, None) for s in shift)
        slice_b = tuple(slice(s, None) for s in shift)
        correlations.append( CF(image[slice_a], image[slice_b]) )

    # return median result
    return np.median(correlations)


def blockwise_cell_quality_score(
    image,
    spacing,
    bounds,
    radius,
    mask=None,
    **kwargs,
):
    """
    """

    # determine stride in voxels
    stride = np.round(radius/spacing).astype(int)

    # define weights for linear blend averaging
    ndim = len(image.shape)
    core = np.array([1.]).reshape((1,)*ndim)
    pad = tuple([s, s] for s in stride)
    weights = np.pad(core, pad, mode='linear_ramp')

    # pad the array to prevent edge effects in averaging
    image = np.pad(image, pad, mode='constant')
    mask = np.pad(mask, pad, mode='constant')

    # get valid sample point coordinates
    samples = np.zeros(image.shape, dtype=bool)
    sample_grid = tuple(slice(s, -s, s) for s in stride)
    samples[sample_grid] = 1
    if mask is not None: samples = samples * mask
    samples = np.nonzero(samples)

    # get container to hold scores
    scores = np.zeros(image.shape, dtype=np.float32)

    # score all blocks
    for iii, coordinate in enumerate(zip(samples[0], samples[1], samples[2])):
        if iii % 100 == 0: print("{} percent complete".format(iii/len(samples[0])))
        context = tuple(slice(x-r, x+r+1) for x, r in zip(coordinate, stride))
        score = cell_quality_score(image[context], spacing, bounds, **kwargs)
        scores[context] += score * weights

    # return result
    crop = tuple(slice(s, -s) for s in stride)
    return scores[crop]


def jaccard_filter(
    fix_mask,
    mov_mask,
    threshold,
):
    """
    """

    fix_mask = fix_mask > 0
    mov_mask = mov_mask > 0
    score = np.sum(fix_mask * mov_mask) / np.sum(fix_mask + mov_mask)
    return True if score >= threshold else False




import SimpleITK as sitk
import h5py
from glob import glob
from os import path
import numpy as np
import ClusterWrap

import dask.bag as db
import dask.array as da

from dask.delayed import delayed


def testPathExtensionForHDF5(image_path):
    """
    Returns true if `image_path` has an hdf5 like extension
    Currently: `['.h5', '.hdf5']
    """

    hdf5_extensions = ['.h5', '.hdf5']
    if image_path in hdf5_extensions:
        return True
    else:
        ext = path.splitext(image_path)[-1]
        return ext in hdf5_extensions


def testPathExtensionForSTACK(image_path):
    """
    Returns true if `image_path` has a stack extension
    Currently: ['.stack']
    """

    stack_extensions = ['.stack',]
    if image_path in stack_extensions:
        return True
    else:
        ext = path.splitext(image_path)[-1]
        return ext in stack_extensions


def globPaths(folder, prefix, suffix):
    """
    Returns sorted list of all absolue paths matching `folder/prefix*suffix`
    If no such paths are found, throws AsserionError
    """

    folder = path.abspath(folder)
    images = sorted(glob(path.join(folder, prefix) + '*' + suffix))
    assert (len(images) > 0), f"No images in {folder} matching {prefix}*{suffix}"
    return images


def daskBagOfFilePaths(folder, prefix, suffix, npartitions=None):
    """
    Returns dask.bag of absolute paths matching `folder/prefix*suffix`
    Specify bag partitions with `npartitions`; default None
    if `npartitions == None` then each filepath is its own partition
    """

    images = globPaths(folder, prefix, suffix)
    if npartitions is None: npartitions = len(images)
    return db.from_sequence(images, npartitions=npartitions)


def daskArrayBackedByHDF5(folder, prefix, suffix, dataset_path, stride=None):
    """
    Returns dask.array backed by HDF5 files matching absolute path `folder/prefix*suffix`
    You must specify hdf5 dataset path with `dataset_path`
    """

    error_message = "daskArrayBackedByHDF5 requires hdf5 files with .h5 or .hdf5 extension"
    assert (testPathExtensionForHDF5(suffix)), error_message
    images = globPaths(folder, prefix, suffix)
    if stride is not None:
        images = images[::stride]

    ex = readHDF5(images[0], dataset_path)
    readHDF5_d = lambda img: da.from_delayed(
        delayed(readHDF5)(img, dataset_path), ex.shape, ex.dtype
    )
    arrays = [readHDF5_d(image) for image in images]
    return da.stack(arrays, axis=0)


def daskArrayBackedBySTACK(folder, prefix, suffix, dtype, shape, stride=None):
    """
    Returns dask.array backed by raw STACK files matching absolute path
    `folder/prefix*suffix`.
    """

    error_message = "daskArrayBackedBySTACK requires stack files with .stack extension"
    assert (testPathExtensionForSTACK(suffix)), error_message
    images = globPaths(folder, prefix, suffix)
    if stride is not None:
        images = images[::stride]

    readSTACK_d = lambda img: da.from_delayed(
        delayed(readSTACK)(img, dtype, shape), shape, dtype,
    )
    arrays = [readSTACK_d(image) for image in images]
    return da.stack(arrays, axis=0)


def readHDF5(image_path, dataset_path):
    """
    Returns array to dataset at `image_path[dataset_path]`
    """

    error_message = "readHDF5 requires hdf5 files with .h5 or .hdf5 extension"
    assert (testPathExtensionForHDF5(image_path)), error_message
    return h5py.File(image_path, 'r')[dataset_path][:]


def readSTACK(image_path, dtype, shape):
    """
    Returns array to dataset at image_path with given data type and shape
    """

    error_message = "readSTACK required stack files with .stack extension"
    assert (testPathExtensionForSTACK(image_path)), error_message
    return np.fromfile(image_path, dtype=dtype).reshape(shape)


def readImage(image_path, dataset_path=None):
    """
    Returns array like object for dataset at `image_path`
    If `image_path` is an hdf5 file, you must specify `dataset_path`
    If `image_path` is an hdf5 file, the return array is a lazy h5py File object
    Else, `image_path` is read by SimpleITK and an in memory numpy array is returned
    File format must be either hdf5 or supported by SimpleITK file readers
    """

    if testPathExtensionForHDF5(image_path):
        assert(dataset_path is not None), "Must provide dataset_path for .h5/.hdf5 files"
        return readHDF5(image_path, dataset_path)
    else:
        return sitk.GetArrayFromImage(sitk.ReadImage(image_path))


def writeHDF5(image_path, dataset_path, array):
    """
    Writes `array` to `image_path[dataset_path]` as an hdf5 file
    """

    with h5py.File(image_path, 'w') as f:
        dset = f.create_dataset(dataset_path, array.shape, array.dtype)
        dset[...] = array


def writeImage(image_path, array, spacing=None, axis_order='zyx'):
    """
    Writes `array` to `image_path`
    `image_path` extension determines format - must be supported by SimpleITK image writers
    Many formats support voxel spacing in the metadata, to specify set `spacing`
    If format supports voxel spacing in meta data, you can set with `spacing`
    Use `axis_order` to specify current axis ordering; default: 'zyx'
    """

    # argsort axis_order
    transpose_order = [i for (v, i) in sorted((v, i) for (i, v) in enumerate(axis_order))]
    array = array.transpose(transpose_order[::-1])  # sitk inverts axis order
    img = sitk.GetImageFromArray(array)
    if spacing is not None:
        spacing = spacing[transpose_order]
        img.SetSpacing(spacing[::-1])
    sitk.WriteImage(img, image_path)


def ensureArray(reference, dataset_path):
    """
    """

    if not isinstance(reference, np.ndarray):
        if not isinstance(reference, str):
            raise ValueError("image references must be ndarrays or filepaths")
        reference = readImage(reference, dataset_path)[...]  # hdf5 arrays are lazy
    return reference


def stack_to_hdf5(stack_path, write_path, dims, dtype):
    """
    """

    stack = np.fromfile(stack_path, dtype=dtype).reshape(dims)
    writeHDF5(write_path, '/default', stack)


def distributed_stack_to_hdf5(
    folder, prefix, suffix,
    dims, dtype,
    cluster_kwargs={}
):
    """
    """

    stack_paths = globPaths(folder, prefix, suffix)
    write_paths = [path.splitext(s)[0] + '.h5' for s in stack_paths]
    nimages = len(stack_paths)
    stack_paths_b = db.from_sequence(stack_paths, npartitions=nimages)
    write_paths_b = db.from_sequence(write_paths, npartitions=nimages)
    with ClusterWrap.cluster(**cluster_kwargs) as cluster:
        stack_paths_b.map(stack_to_hdf5, write_paths_b, dims, dtype).compute()




import numpy as np
import SimpleITK as sitk
import functools
import ClusterWrap
import dask
from scipy.ndimage import zoom


def skip_sample(image, spacing, ss_spacing):
    """
    """

    ss = np.maximum(np.round(ss_spacing / spacing), 1).astype(np.int)
    image = image[::ss[0], ::ss[1], ::ss[2]]
    spacing = spacing * ss
    return image, spacing


def numpy_to_sitk(image, spacing, origin=None, vector=False):
    """
    """

    # check endianness of data - some sitk operations seem to
    # only work with little endian
    if str(image.dtype)[0] == '>':
        error = "Array cannot be big endian. Convert arrays with ndarray.astype\n"
        error += "Given array dtype is " + str(image.dtype)
        raise TypeError(error)

    image = sitk.GetImageFromArray(image.copy(), isVector=vector)
    image.SetSpacing(spacing[::-1])
    if origin is None:
        origin = np.zeros(len(spacing))
    image.SetOrigin(origin[::-1])
    return image


def invert_matrix_axes(matrix):
    """
    """

    corrected = np.eye(4)
    corrected[:3, :3] = matrix[:3, :3][::-1, ::-1]
    corrected[:3, -1] = matrix[:3, -1][::-1]
    return corrected


def affine_transform_to_matrix(transform):
    """
    """

    matrix = np.eye(4)
    matrix[:3, :3] = np.array(transform.GetMatrix()).reshape((3,3))
    matrix[:3, -1] = np.array(transform.GetTranslation())
    return invert_matrix_axes(matrix)


def matrix_to_affine_transform(matrix):
    """
    """

    matrix_sitk = invert_matrix_axes(matrix)
    transform = sitk.AffineTransform(3)
    transform.SetMatrix(matrix_sitk[:3, :3].flatten())
    transform.SetTranslation(matrix_sitk[:3, -1].squeeze())
    return transform


def matrix_to_euler_transform(matrix):
    """
    """

    matrix_sitk = invert_matrix_axes(matrix)
    transform = sitk.Euler3DTransform()
    transform.SetMatrix(matrix_sitk[:3, :3].flatten())
    transform.SetTranslation(matrix_sitk[:3, -1].squeeze())
    return transform


def euler_transform_to_parameters(transform):
    """
    """

    return np.array((transform.GetAngleX(),
                     transform.GetAngleY(),
                     transform.GetAngleZ()) +
                     transform.GetTranslation()
    )


def parameters_to_euler_transform(params):
    """
    """

    transform = sitk.Euler3DTransform()
    transform.SetRotation(*params[:3])
    transform.SetTranslation(params[3:])
    return transform


def matrix_to_displacement_field(reference, matrix, spacing):
    """
    """

    nrows, ncols, nstacks = reference.shape
    grid = np.array(np.mgrid[:nrows, :ncols, :nstacks]).transpose(1,2,3,0)
    grid = grid * spacing
    mm, tt = matrix[:3, :3], matrix[:3, -1]
    return np.einsum('...ij,...j->...i', mm, grid) + tt - grid


def field_to_displacement_field_transform(field, spacing):
    """
    """

    field = field.astype(np.float64)[..., ::-1]
    transform = numpy_to_sitk(field, spacing, vector=True)
    return sitk.DisplacementFieldTransform(transform)


def bspline_parameters_to_transform(parameters):
    """
    """

    t = sitk.BSplineTransform(3, 3)
    t.SetFixedParameters(parameters[:18])
    t.SetParameters(parameters[18:])
    return t


def bspline_to_displacement_field(reference, bspline, shape=None):
    """
    """

    df = sitk.TransformToDisplacementField(
        bspline, sitk.sitkVectorFloat64,
        reference.GetSize(), reference.GetOrigin(),
        reference.GetSpacing(), reference.GetDirection(),
    )
    df = sitk.GetArrayFromImage(df).astype(np.float32)[..., ::-1]
    if shape is not None:
        new_shape = [x / y for x, y in zip(shape, df.shape[:-1])] + [1,]
        df = zoom(df, new_shape, order=1, mode='nearest')
    return df


def scatter_dask_array(cluster, array):
    """
    """

    if not isinstance(array, dask.array.Array):
        future = cluster.client.scatter(array)
        da = dask.array.from_delayed(future, shape=array.shape, dtype=array.dtype)
        return da
    else:
        return array




import sys, time
import numpy as np
import dask.array as da
import SimpleITK as sitk
from ClusterWrap.decorator import cluster
import CircuitSeeker.utility as ut
from CircuitSeeker._wrap_irm import configure_irm
from CircuitSeeker.transform import apply_transform
from CircuitSeeker.transform import compose_affine_and_displacement_vector_field
from CircuitSeeker.transform import compose_displacement_vector_fields
from CircuitSeeker.quality import jaccard_filter
from CircuitSeeker.metrics import patch_mutual_information
from scipy.spatial.transform import Rotation


def random_affine_search(
    fix,
    mov,
    fix_spacing,
    mov_spacing,
    max_translation,
    max_rotation,
    max_scale,
    max_shear,
    random_iterations,
    affine_align_best=0,
    alignment_spacing=None,
    fix_mask=None,
    mov_mask=None,
    fix_origin=None,
    mov_origin=None,
    jaccard_filter_threshold=None,
    use_patch_mutual_information=False,
    print_running_improvements=False,
    **kwargs,
):
    """
    Apply random affine matrices within given bounds to moving image. The best
    scoring affines can be further refined with gradient descent based affine
    alignment. The single best result is returned. This function is intended
    to find good initialization for a full affine alignment obtained by calling
    `affine_align`

    Parameters
    ----------
    fix : ndarray
        the fixed image

    mov : ndarray
        the moving image; `fix.ndim` must equal `mov.ndim`

    fix_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the fixed image.
        Length must equal `fix.ndim`

    mov_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the moving image.
        Length must equal `mov.ndim`

    max_translation : float
        The maximum amplitude translation allowed in random sampling.
        Specified in physical units (e.g. um or mm)

    max_rotation : float
        The maximum amplitude rotation allowed in random sampling.
        Specified in radians

    max_scale : float
        The maximum amplitude scaling allowed in random sampling.

    max_shear : float
        The maximum amplitude shearing allowed in random sampling.

    random_iterations : int
        The number of random affine matrices to sample

    affine_align_best : int (default: 0)
        The best `affine_align_best` random affine matrices are refined
        by calling `affine_align` setting the random affine as the
        `initial_transform`. This is parameterized through **kwargs.

    alignment_spacing : float (default: None)
        Fixed and moving images are skip sampled to a voxel spacing
        as close as possible to this value. Intended for very fast
        simple alignments (e.g. low amplitude motion correction)

    fix_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the fixed image

    mov_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the moving image

    fix_origin : 1d array (default: None)
        Origin of the fixed image.
        Length must equal `fix.ndim`

    mov_origin : 1d array (default: None)
        Origin of the moving image.
        Length must equal `mov.ndim`

    jaccard_filter_threshold : float in range [0, 1] (default: None)
        If `jaccard_filter_threshold`, `fix_mask`, and `mov_mask` are all
        defined (i.e. not None), then the Jaccard index between the masks
        is computed. If the index is less than this threshold the alignment
        is skipped and the default is returned. Useful for distributed piecewise
        workflows over heterogenous data.

    **kwargs : any additional arguments
        Passed to `configure_irm` to score random affines
        Also passed to `affine_align` for gradient descent
        based refinement

    Returns
    -------
    transform : 4x4 array
        The (refined) random affine matrix best initializing a match of
        the moving image to the fixed. Should be further refined by calling
        `affine_align`.
    """

    # check jaccard index
    a = jaccard_filter_threshold is not None
    b = fix_mask is not None
    c = mov_mask is not None
    if a and b and c:
        if not jaccard_filter(fix_mask, mov_mask, jaccard_filter_threshold):
            print("Masks failed jaccard_filter")
            print("Returning default")
            return np.eye(4)

    # define conversion from params to affine transform
    def params_to_affine_matrix(params):

        # translation
        translation = np.eye(4)
        translation[:3, -1] = params[:3]

        # rotation
        rotation = np.eye(4)
        rotation[:3, :3] = Rotation.from_rotvec(params[3:6]).as_matrix()
        center = np.array(fix.shape) / 2 * fix_spacing
        tl, tr = np.eye(4), np.eye(4)
        tl[:3, -1], tr[:3, -1] = center, -center
        rotation = np.matmul(tl, np.matmul(rotation, tr))

        # scale
        scale = np.diag( list(params[6:9]) + [1,])

        # shear
        shx, shy, shz = np.eye(4), np.eye(4), np.eye(4)
        shx[1, 0], shx[2, 0] = params[10], params[11]
        shy[0, 1], shy[2, 1] = params[9], params[11]
        shz[0, 2], shz[1, 2] = params[9], params[10]
        shear = np.matmul(shz, np.matmul(shy, shx))

        # compose
        aff = np.matmul(rotation, translation)
        aff = np.matmul(scale, aff)
        aff = np.matmul(shear, aff)
        return aff
        
    # generate random parameters, first row is always identity
    params = np.zeros((random_iterations+1, 12))
    params[:, 6:9] = 1  # default for scale params
    F = lambda mx: 2 * mx * np.random.rand(random_iterations, 3) - mx
    if max_translation != 0: params[1:, 0:3] = F(max_translation)
    if max_rotation != 0: params[1:, 3:6] = F(max_rotation)
    if max_scale != 1: params[1:, 6:9] = np.e**F(np.log(max_scale))
    if max_shear != 0: params[1:, 9:] = F(max_shear)

    # skip sample to alignment spacing
    if alignment_spacing is not None:
        fix, fix_spacing_ss = ut.skip_sample(fix, fix_spacing, alignment_spacing)
        mov, mov_spacing_ss = ut.skip_sample(mov, mov_spacing, alignment_spacing)
        if fix_mask is not None:
            fix_mask, _ = ut.skip_sample(fix_mask, fix_spacing, alignment_spacing)
        if mov_mask is not None:
            mov_mask, _ = ut.skip_sample(mov_mask, mov_spacing, alignment_spacing)
        fix_spacing = fix_spacing_ss
        mov_spacing = mov_spacing_ss

    # define metric evaluation
    if use_patch_mutual_information:

        # wrap patch_mi metric
        def score_affine(affine):

            # apply transform
            aligned = apply_transform(
                fix, mov, fix_spacing, mov_spacing,
                transform_list=[affine,],
                fix_origin=fix_origin,
                mov_origin=mov_origin,
            )

            # mov mask
            mov_mask_aligned = None
            if mov_mask is not None:
                mov_mask_aligned = apply_transform(
                    fix, mov_mask, fix_spacing, mov_spacing,
                    transform_list=[affine,],
                    fix_origin=fix_origin,
                    mov_origin=mov_origin,
                    interpolate_with_nn=True,
                )

            return patch_mutual_information(
                fix, aligned, fix_spacing,
                fix_mask=fix_mask,
                mov_mask=mov_mask_aligned,
                return_metric_image=False,
                **kwargs,
            )

    # otherwise score entire image domain
    else:

        # convert to float32 sitk images
        fix_sitk = ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin)
        mov_sitk = ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin)
        fix_sitk = sitk.Cast(fix_sitk, sitk.sitkFloat32)
        mov_sitk = sitk.Cast(mov_sitk, sitk.sitkFloat32)

        # create irm to use metric object
        irm = configure_irm(**kwargs)

        # set masks
        if fix_mask is not None:
            fix_mask_sitk = ut.numpy_to_sitk(fix_mask, fix_spacing, origin=fix_origin)
            irm.SetMetricFixedMask(fix_mask_sitk)
        if mov_mask is not None:
            mov_mask_sitk = ut.numpy_to_sitk(mov_mask, mov_spacing, origin=mov_origin)
            irm.SetMetricMovingMask(mov_mask_sitk)

        # wrap full image mi
        def score_affine(affine):

            # reformat affine as tranfsorm and give to irm
            affine = ut.matrix_to_affine_transform(affine)
            irm.SetMovingInitialTransform(affine)

            # get the metric value
            try:
                return irm.MetricEvaluate(fix_sitk, mov_sitk)
            except Exception as e:
                return np.finfo(scores.dtype).max

    # score all random affines
    fail_count = 0
    current_best_score = 0
    scores = np.empty(random_iterations + 1)
    for iii, ppp in enumerate(params):
        aff = params_to_affine_matrix(ppp)
        scores[iii] = score_affine(aff)
        if scores[iii] == np.finfo(scores.dtype).max:
            fail_count += 1

        # print running improvements
        if print_running_improvements:
            if scores[iii] < current_best_score:
                current_best_score = scores[iii]
                print(iii, ': ', current_best_score, '\n', aff, '\n', flush=True)

        # check for excessive failure
        if fail_count >= 10 or fail_count >= random_iterations + 1:
            print("Random search failed due to ITK exceptions")
            print("Returning default")
            return np.eye(4)

    # sort
    params = params[np.argsort(scores)]

    # gradient descent based refinements
    if affine_align_best == 0:
        return params_to_affine_matrix(params[0])

    else:
        # container to hold the scores
        scores = np.empty(affine_align_best)
        fail_count = 0  # keep track of failures
        for iii in range(affine_align_best):

            # gradient descent affine alignment
            aff = params_to_affine_matrix(params[iii])
            aff = affine_align(
               fix, mov, fix_spacing, mov_spacing,
               initial_transform=aff,
               fix_mask=fix_mask,
               mov_mask=mov_mask,
               fix_origin=fix_origin,
               mov_origin=mov_origin,
               alignment_spacing=None,  # already done in this function
               **kwargs,
            )

            # score the result
            scores[iii] = score_affine(aff)
            if fail_count >= affine_align_best:
                print("Random search failed due to ITK exceptions")
                print("Returning default")
                return np.eye(4)

        # return the best one
        return params_to_affine_matrix(params[np.argmin(scores)])
        

def affine_align(
    fix,
    mov,
    fix_spacing,
    mov_spacing,
    rigid=False,
    initial_transform=None,
    initialize_with_centering=False,
    alignment_spacing=None,
    fix_mask=None,
    mov_mask=None,
    fix_origin=None,
    mov_origin=None,
    jaccard_filter_threshold=None,
    default=np.eye(4),
    **kwargs,
):
    """
    Affine or rigid alignment of a fixed/moving image pair.
    Lots of flexibility in speed/accuracy trade off.
    Highly configurable and useful in many contexts.

    Parameters
    ----------
    fix : ndarray
        the fixed image

    mov : ndarray
        the moving image; `fix.ndim` must equal `mov.ndim`

    fix_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the fixed image.
        Length must equal `fix.ndim`

    mov_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the moving image.
        Length must equal `mov.ndim`

    rigid : bool (default: False)
        Restrict the alignment to rigid motion only

    initial_transform : 4x4 array (default: None)
        An initial rigid or affine matrix from which to initialize
        the optimization

    initialize_with_center : bool (default: False)
        Initialize the optimization center of mass translation
        Cannot be True if `initial_transform` is not None

    alignment_spacing : float (default: None)
        Fixed and moving images are skip sampled to a voxel spacing
        as close as possible to this value. Intended for very fast
        simple alignments (e.g. low amplitude motion correction)

    fix_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the fixed image

    mov_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the moving image

    fix_origin : 1d array (default: None)
        Origin of the fixed image.
        Length must equal `fix.ndim`

    mov_origin : 1d array (default: None)
        Origin of the moving image.
        Length must equal `mov.ndim`

    jaccard_filter_threshold : float in range [0, 1] (default: None)
        If `jaccard_filter_threshold`, `fix_mask`, and `mov_mask` are all
        defined (i.e. not None), then the Jaccard index between the masks
        is computed. If the index is less than this threshold the alignment
        is skipped and the default is returned. Useful for distributed piecewise
        workflows over heterogenous data.

    default : 4x4 array (default: identity matrix)
        If the optimization fails, print error message but return this value

    **kwargs : any additional arguments
        Passed to `configure_irm`
        This is where you would set things like:
        metric, iterations, shrink_factors, and smooth_sigmas

    Returns
    -------
    transform : 4x4 array
        The affine or rigid transform matrix matching moving to fixed
    """

    # update default if an initial transform is provided
    if initial_transform is not None and np.all(default == np.eye(4)):
        default = initial_transform

    # check jaccard index
    a = jaccard_filter_threshold is not None
    b = fix_mask is not None
    c = mov_mask is not None
    if a and b and c:
        if not jaccard_filter(fix_mask, mov_mask, jaccard_filter_threshold):
            print("Masks failed jaccard_filter")
            print("Returning default")
            return default

    # skip sample to alignment spacing
    if alignment_spacing is not None:
        fix, fix_spacing_ss = ut.skip_sample(fix, fix_spacing, alignment_spacing)
        mov, mov_spacing_ss = ut.skip_sample(mov, mov_spacing, alignment_spacing)
        if fix_mask is not None:
            fix_mask, _ = ut.skip_sample(fix_mask, fix_spacing, alignment_spacing)
        if mov_mask is not None:
            mov_mask, _ = ut.skip_sample(mov_mask, mov_spacing, alignment_spacing)
        fix_spacing = fix_spacing_ss
        mov_spacing = mov_spacing_ss

    # convert to float32 sitk images
    fix = ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin)
    mov = ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin)
    fix = sitk.Cast(fix, sitk.sitkFloat32)
    mov = sitk.Cast(mov, sitk.sitkFloat32)

    # set up registration object
    irm = configure_irm(**kwargs)

    # select initial transform type
    if rigid and initial_transform is None:
        transform = sitk.Euler3DTransform()
    elif rigid and initial_transform is not None:
        transform = ut.matrix_to_euler_transform(initial_transform)
    elif not rigid and initial_transform is None:
        transform = sitk.AffineTransform(3)
    elif not rigid and initial_transform is not None:
        transform = ut.matrix_to_affine_transform(initial_transform)

    # consider initializing with centering
    if initial_transform is None and initialize_with_centering:
        transform = sitk.CenteredTransformInitializer(
            fix, mov, transform,
        )

    # set initial transform
    irm.SetInitialTransform(transform, inPlace=True)

    # set masks
    if fix_mask is not None:
        fix_mask = ut.numpy_to_sitk(fix_mask, fix_spacing, origin=fix_origin)
        irm.SetMetricFixedMask(fix_mask)
    if mov_mask is not None:
        mov_mask = ut.numpy_to_sitk(mov_mask, mov_spacing, origin=mov_origin)
        irm.SetMetricMovingMask(mov_mask)

    # execute alignment, for any exceptions return default
    try:
        initial_metric_value = irm.MetricEvaluate(fix, mov)
        irm.Execute(fix, mov)
        final_metric_value = irm.MetricEvaluate(fix, mov)
    except Exception as e:
        print("Registration failed due to ITK exception:\n", e)
        print("\nReturning default")
        sys.stdout.flush()
        return default

    # if centered, convert back to Euler3DTransform object
    if rigid and initialize_with_centering:
        transform = sitk.Euler3DTransform(transform)

    # if registration improved metric return result
    # otherwise return default
    if final_metric_value < initial_metric_value:
        sys.stdout.flush()
        return ut.affine_transform_to_matrix(transform)
    else:
        print("Optimization failed to improve metric")
        print("initial value: {}".format(initial_metric_value))
        print("final value: {}".format(final_metric_value))
        print("Returning default")
        sys.stdout.flush()
        return default


def deformable_align(
    fix,
    mov,
    fix_spacing,
    mov_spacing,
    control_point_spacing,
    control_point_levels,
    initial_transform=None,
    alignment_spacing=None,
    fix_mask=None,
    mov_mask=None,
    fix_origin=None,
    mov_origin=None,
    jaccard_filter_threshold=None,
    default=None,
    **kwargs,
):
    """
    Register moving to fixed image with a bspline parameterized deformation field

    Parameters
    ----------
    fix : ndarray
        the fixed image

    mov : ndarray
        the moving image; `fix.ndim` must equal `mov.ndim`

    fix_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the fixed image.
        Length must equal `fix.ndim`

    mov_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the moving image.

    control_point_spacing : float
        The spacing in physical units (e.g. mm or um) between control
        points that parameterize the deformation. Smaller means
        more precise alignment, but also longer compute time. Larger
        means shorter compute time and smoother transform, but less
        precise.

    control_point_levels : list of type int
        The optimization scales for control point spacing. E.g. if
        `control_point_spacing` is 100.0 and `control_point_levels`
        is [1, 2, 4] then method will optimize at 400.0 units control
        points spacing, then optimize again at 200.0 units, then again
        at the requested 100.0 units control point spacing.
    
    initial_transform : 4x4 array (default: None)
        An initial rigid or affine matrix from which to initialize
        the optimization

    alignment_spacing : float (default: None)
        Fixed and moving images are skip sampled to a voxel spacing
        as close as possible to this value. Intended for very fast
        simple alignments (e.g. low amplitude motion correction)

    fix_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the fixed image

    mov_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the moving image

    fix_origin : 1d array (default: None)
        Origin of the fixed image.
        Length must equal `fix.ndim`

    mov_origin : 1d array (default: None)
        Origin of the moving image.
        Length must equal `mov.ndim`

    jaccard_filter_threshold : float in range [0, 1] (default: None)
        If `jaccard_filter_threshold`, `fix_mask`, and `mov_mask` are all
        defined (i.e. not None), then the Jaccard index between the masks
        is computed. If the index is less than this threshold the alignment
        is skipped and the default is returned. Useful for distributed piecewise
        workflows over heterogenous data.

    default : any object (default: None)
        If optimization fails to improve image matching metric,
        print an error but also return this object. If None
        the parameters and displacement field for an identity
        transform are returned.

    **kwargs : any additional arguments
        Passed to `configure_irm`
        This is where you would set things like:
        metric, iterations, shrink_factors, and smooth_sigmas

    Returns
    -------
    params : 1d array
        The complete set of control point parameters concatenated
        as a 1d array.

    field : ndarray
        The displacement field parameterized by the bspline control
        points
    """

    # check jaccard index
    a = jaccard_filter_threshold is not None
    b = fix_mask is not None
    c = mov_mask is not None
    failed_jaccard = False
    if a and b and c:
        if not jaccard_filter(fix_mask, mov_mask, jaccard_filter_threshold):
            print("Masks failed jaccard_filter")
            print("Returning default")
            failed_jaccard = True

    # store initial fixed image shape
    initial_fix_shape = fix.shape

    # skip sample to alignment spacing
    if alignment_spacing is not None:
        fix, fix_spacing_ss = ut.skip_sample(fix, fix_spacing, alignment_spacing)
        mov, mov_spacing_ss = ut.skip_sample(mov, mov_spacing, alignment_spacing)
        if fix_mask is not None:
            fix_mask, _ = ut.skip_sample(fix_mask, fix_spacing, alignment_spacing)
        if mov_mask is not None:
            mov_mask, _ = ut.skip_sample(mov_mask, mov_spacing, alignment_spacing)
        fix_spacing = fix_spacing_ss
        mov_spacing = mov_spacing_ss

    # convert to sitk images, float32 type
    fix = ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin)
    mov = ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin)
    fix = sitk.Cast(fix, sitk.sitkFloat32)
    mov = sitk.Cast(mov, sitk.sitkFloat32)

    # set up registration object
    irm = configure_irm(**kwargs)

    # set initial moving transform
    if initial_transform is not None:
        if len(initial_transform.shape) == 2:
            it = ut.matrix_to_affine_transform(initial_transform)
        irm.SetMovingInitialTransform(it)

    # get control point grid shape
    fix_size_physical = [sz*sp for sz, sp in zip(fix.GetSize(), fix.GetSpacing())]
    x, y = control_point_spacing, control_point_levels[-1]
    control_point_grid = [max(1, int(sz / (x*y))) for sz in fix_size_physical]

    # set initial transform
    transform = sitk.BSplineTransformInitializer(
        image1=fix, transformDomainMeshSize=control_point_grid, order=3,
    )
    irm.SetInitialTransformAsBSpline(
        transform, inPlace=True, scaleFactors=control_point_levels,
    )

    # store initial transform coordinates as default
    if default is None:
        fp = transform.GetFixedParameters()
        pp = transform.GetParameters()
        default_params = np.array(list(fp) + list(pp))
        default_field = ut.bspline_to_displacement_field(
            fix, transform, shape=initial_fix_shape,
        )
        default = (default_params, default_field)

    # now that default is defined, determine jaccard result
    if failed_jaccard: return default

    # set masks
    if fix_mask is not None:
        fix_mask = ut.numpy_to_sitk(fix_mask, fix_spacing, origin=fix_origin)
        irm.SetMetricFixedMask(fix_mask)
    if mov_mask is not None:
        mov_mask = ut.numpy_to_sitk(mov_mask, mov_spacing, origin=mov_origin)
        irm.SetMetricMovingMask(mov_mask)

    # execute alignment, for any exceptions return default
    try:
        initial_metric_value = irm.MetricEvaluate(fix, mov)
        irm.Execute(fix, mov)
        final_metric_value = irm.MetricEvaluate(fix, mov)
    except Exception as e:
        print("Registration failed due to ITK exception:\n", e)
        print("\nReturning default")
        sys.stdout.flush()
        return default

    # if registration improved metric return result
    # otherwise return default
    if final_metric_value < initial_metric_value:
        sys.stdout.flush()
        fp = transform.GetFixedParameters()
        pp = transform.GetParameters()
        params = np.array(list(fp) + list(pp))
        field = ut.bspline_to_displacement_field(
            fix, transform, shape=initial_fix_shape,
        )
        return params, field
    else:
        print("Optimization failed to improve metric")
        print("initial value: {}".format(initial_metric_value))
        print("final value: {}".format(final_metric_value))
        print("Returning default")
        sys.stdout.flush()
        return default


def alignment_pipeline(
    fix,
    mov,
    fix_spacing,
    mov_spacing,
    steps,
    initial_transform=None,
    fix_mask=None,
    mov_mask=None,
    fix_origin=None,
    mov_origin=None,
    random_kwargs={},
    rigid_kwargs={},
    affine_kwargs={},
    deform_kwargs={},
    **kwargs,
):
    """
    Compose random, rigid, affine, and deformable alignment with one function call

    Parameters
    ----------
    fix : ndarray
        the fixed image

    mov : ndarray
        the moving image; `fix.ndim` must equal `mov.ndim`

    fix_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the fixed image.
        Length must equal `fix.ndim`

    mov_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the moving image.

    steps : list of strings
        Alignment steps to run. Options include:
        'random' : run `random_affine_search`
        'rigid' : run `affine_align` with `rigid=True`
        'affine' : run `affine_align`
        'deform' : run `deformable_align`

        Currently you cannot run 'random' and 'rigid' in the same pipeline.
        Hoping to enable this in the future.

    initial_transform : ndarray (default: None)
        An initial transform. This should be a 4x4 affine matrix.
        Not compatible with 'random' in `steps`

    fix_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the fixed image

    mov_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the moving image

    fix_origin : 1d array (default: None)
        Origin of the fixed image.
        Length must equal `fix.ndim`

    mov_origin : 1d array (default: None)
        Origin of the moving image.
        Length must equal `mov.ndim`

    random_kwargs : dict (default: None)
        Arguments passed to `random_affine_search`

    rigid_kwargs : dict (default: None)
        Arguments passed to `affine_align` when `rigid=True` (rigid alignment)

    affine_kwargs : dict (default: None)
        Arguments passed to `affine_align` when `rigid=False` (affine alignment)

    deform_kwargs : dict (default: None)
        Arguments passed to `deformable_align`

    **kwargs : any additional keyword arguments
        Global arguments that apply to all alignment steps
        These are overwritten by specific arguments passed via
        `random_kwargs`, `rigid_kwargs`, `affine_kwargs`, and
        `deform_kwargs`

    Returns
    -------
    transform : ndarray or tuple of ndarray
        Transform(s) aligning moving to fixed image.

        If 'deform' is not in `steps` then this is a single 4x4 matrix - all
        steps ('random', 'rigid', and/or 'affine') are composed.

        If 'deform' is in `steps` then this is a tuple. The first element
        is the composed 4x4 affine matrix, the second is a displacement
        vector field with shape equal to fix.shape + (3,)
    """

    # TODO: rigid could be done before random, but random would
    #       need to support an initial_transform; possible
    #       with matrix multiplication (composition)
    if 'random' in steps and 'rigid' in steps:
        message = "cannot do rigid alignment after random affine\n"
        message += "remove either 'random' or 'rigid' from steps"
        raise ValueError(message)

    # set default
    affine = initial_transform if initial_transform is not None else np.eye(4)

    # establish all keyword arguments
    random_kwargs = {**kwargs, **random_kwargs}
    rigid_kwargs = {**kwargs, **rigid_kwargs}
    affine_kwargs = {**kwargs, **affine_kwargs}
    deform_kwargs = {**kwargs, **deform_kwargs}

    # random initialization
    if 'random' in steps:
        affine = random_affine_search(
            fix, mov,
            fix_spacing, mov_spacing,
            fix_mask=fix_mask,
            mov_mask=mov_mask,
            fix_origin=fix_origin,
            mov_origin=mov_origin,
            **random_kwargs,
         )
    # rigid alignment
    if 'rigid' in steps:
        affine = affine_align(
            fix, mov,
            fix_spacing, mov_spacing,
            fix_mask=fix_mask,
            mov_mask=mov_mask,
            fix_origin=fix_origin,
            mov_origin=mov_origin,
            initial_transform=affine,
            rigid=True,
            **rigid_kwargs,
        )
    # affine alignment
    if 'affine' in steps:
        affine = affine_align(
            fix, mov,
            fix_spacing, mov_spacing,
            fix_mask=fix_mask,
            mov_mask=mov_mask,
            fix_origin=fix_origin,
            mov_origin=mov_origin,
            initial_transform=affine,
            **affine_kwargs,
        )
    # deformable align
    if 'deform' in steps:
        deform = deformable_align(
            fix, mov,
            fix_spacing, mov_spacing,
            fix_mask=fix_mask,
            mov_mask=mov_mask,
            fix_origin=fix_origin,
            mov_origin=mov_origin,
            initial_transform=affine,
            **deform_kwargs,
        )
        return affine, deform

    # return affine result
    else:
        return affine


@cluster
def distributed_piecewise_alignment_pipeline(
    fix,
    mov,
    fix_spacing,
    mov_spacing,
    nblocks,
    overlap=0.5,
    fix_mask=None,
    mov_mask=None,
    steps=['rigid', 'affine'],
    random_kwargs={},
    rigid_kwargs={},
    affine_kwargs={},
    deform_kwargs={},
    cluster=None,
    cluster_kwargs={},
    **kwargs,
):
    """
    Piecewise affine alignment of moving to fixed image.
    Overlapping blocks are given to `affine_align` in parallel
    on distributed hardware. Can include random initialization,
    rigid alignment, and affine alignment.

    Parameters
    ----------
    fix : ndarray
        the fixed image

    mov : ndarray
        the moving image; `fix.shape` must equal `mov.shape`
        I.e. typically piecewise affine alignment is done after
        a global affine alignment wherein the moving image has
        been resampled onto the fixed image voxel grid.

    fix_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the fixed image.
        Length must equal `fix.ndim`

    mov_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the moving image.
        Length must equal `mov.ndim`

    nblocks : iterable
        The number of blocks to use along each axis.
        Length should be equal to `fix.ndim`

    overlap : float in range [0, 1] (default: 0.5)
        Block overlap size as a percentage of block size

    fix_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the fixed image

    mov_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the moving image
        Due to the distribution aspect, if a mov_mask is provided
        you must also provide a fix_mask. A reasonable choice if
        no fix_mask exists is an array of all ones.

    steps : list of type string (default: ['rigid', 'affine'])
        Flags to indicate which steps to run. An empty list will guarantee
        all affines are the identity. Any of the following may be in the list:
            'random': run `random_affine_search` first
            'rigid': run `affine_align` with rigid=True
            'affine': run `affine_align` with rigid=False
        If all steps are present they are run in the order given above.
        Steps share parameters given to kwargs. Parameters for individual
        steps override general settings with `random_kwargs`, `rigid_kwargs`,
        and `affine_kwargs`. If `random` is in the list, `random_kwargs`
        must be defined.

    random_kwargs : dict (default: {})
        Keyword arguments to pass to `random_affine_search`. This is only
        necessary if 'random' is in `steps`. If so, the following keys must
        be given:
                'max_translation'
                'max_rotation'
                'max_scale'
                'max_shear'
                'random_iterations'
        However any argument to `random_affine_search` may be defined. See
        documentation for `random_affine_search` for descriptions of these
        parameters. If 'random' and 'rigid' are both in `steps` then
        'max_scale' and 'max_shear' must both be 0.

    rigid_kwargs : dict (default: {})
        If 'rigid' is in `steps`, these keyword arguments are passed
        to `affine_align` during the rigid=True step. They override
        any common general kwargs.

    affine_kwargs : dict (default: {})
        If 'affine' is in `steps`, these keyword arguments are passed
        to `affine_align` during the rigid=False (affine) step. They
        override any common general kwargs.

    cluster_kwargs : dict (default: {})
        Arguments passed to ClusterWrap.cluster
        If working with an LSF cluster, this will be
        ClusterWrap.janelia_lsf_cluster. If on a workstation
        this will be ClusterWrap.local_cluster.
        This is how distribution parameters are specified.

    kwargs : any additional arguments
        Passed to calls `random_affine_search` and `affine_align` calls

    Returns
    -------
    affines : nd array
        Affine matrix for each block. Shape is (X, Y, ..., 4, 4)
        for X blocks along first axis and so on.

    field : nd array
        Local affines stitched together into a displacement field
        Shape is `fix.shape` + (3,) as the last dimension contains
        the displacement vector.
    """

    # wait for at least one worker to be fully  instantiated
    while ((cluster.client.status == "running") and
           (len(cluster.client.scheduler_info()["workers"]) < 1)):
        time.sleep(1.0)

    # compute block size and overlaps
    blocksize = np.array(fix.shape).astype(np.float32) / nblocks
    blocksize = np.ceil(blocksize).astype(np.int16)
    overlaps = tuple(np.round(blocksize * overlap).astype(np.int16))

    # pad the ends to fill in the last blocks
    # blocks must all be exact for stitch to work correctly
    pads = [(0, y - x % y) if x % y > 0
        else (0, 0) for x, y in zip(fix.shape, blocksize)]
    fix_p = np.pad(fix, pads)
    mov_p = np.pad(mov, pads)

    # pad masks if necessary
    if fix_mask is not None:
        fm_p = np.pad(fix_mask, pads)
    if mov_mask is not None:
        mm_p = np.pad(mov_mask, pads)

    # CONSTRUCT DASK ARRAY VERSION OF OBJECTS
    # fix
    fix_future = cluster.client.scatter(fix_p)
    fix_da = da.from_delayed(
        fix_future, shape=fix_p.shape, dtype=fix_p.dtype
    ).rechunk(tuple(blocksize))

    # mov
    mov_future = cluster.client.scatter(mov_p)
    mov_da = da.from_delayed(
        mov_future, shape=mov_p.shape, dtype=mov_p.dtype
    ).rechunk(tuple(blocksize))

    # fix mask
    if fix_mask is not None:
        fm_future = cluster.client.scatter(fm_p)
        fm_da = da.from_delayed(
            fm_future, shape=fm_p.shape, dtype=fm_p.dtype
        ).rechunk(tuple(blocksize))
    else:
        fm_da = da.from_array([None,], chunks=(1,))

    # mov mask
    if mov_mask is not None:
        mm_future = cluster.client.scatter(mm_p)
        mm_da = da.from_delayed(
            mm_future, shape=mm_p.shape, dtype=mm_p.dtype
        ).rechunk(tuple(blocksize))
    else:
        mm_da = da.from_array([None,], chunks=(1,))

    # establish all keyword arguments
    random_kwargs = {**kwargs, **random_kwargs}
    rigid_kwargs = {**kwargs, **rigid_kwargs}
    affine_kwargs = {**kwargs, **affine_kwargs}
    deform_kwargs = {**kwargs, **deform_kwargs}

    # closure for alignment pipeline
    def align_single_block(fix, mov, fm, mm, block_info=None):
        
        # check for masks
        if fm.shape != fix.shape: fm = None
        if mm.shape != mov.shape: mm = None

        # run alignment pipeline
        transform = alignment_pipeline(
            fix, mov, fix_spacing, mov_spacing, steps,
            fix_mask=fm, mov_mask=mm,
            random_kwargs=random_kwargs,
            rigid_kwargs=rigid_kwargs,
            affine_kwargs=affine_kwargs,
            deform_kwargs=deform_kwargs,
        )

        # convert to single vector field
        if isinstance(transform, tuple):
            affine, deform = transform[0], transform[1][1]
            transform = compose_affine_and_displacement_vector_field(
                affine, deform, fix_spacing,
            )
        else:
            transform = ut.matrix_to_displacement_field(
                fix, transform, fix_spacing,
            )

        # get block index and block grid size
        block_grid = block_info[0]['num-chunks']
        block_index = block_info[0]['chunk-location']

        # create weights array
        core, pad_ones, pad_linear = [], [], []
        for i in range(3):

            # get core shape and pad sizes
            o = max(0, 2*overlaps[i]-1)
            c = blocksize[i] - o + 1
            p_ones, p_linear = [0, 0], [o, o]
            if block_index[i] == 0:
                p_ones[0], p_linear[0] = o//2, 0
            if block_index[i] == block_grid[i] - 1:
                p_ones[1], p_linear[1] = o//2, 0
            core.append(c)
            pad_ones.append(tuple(p_ones))
            pad_linear.append(tuple(p_linear))

        # create weights core
        weights = np.ones(core, dtype=np.float32)

        # extend weights
        weights = np.pad(
            weights, pad_ones, mode='constant', constant_values=1,
        )
        weights = np.pad(
            weights, pad_linear, mode='linear_ramp', end_values=0,
        )

        # apply weights
        transform = transform * weights[..., None]

        # package and return result
        result = np.empty((1,1,1), dtype=object)
        result[0, 0, 0] = transform
        return result
    # END CLOSURE

    # determine overlaps list
    overlaps_list = [overlaps, overlaps]
    if fix_mask is not None:
        overlaps_list.append(overlaps)
    else:
        overlaps_list.append(0)
    if mov_mask is not None:
        overlaps_list.append(overlaps)
    else:
        overlaps_list.append(0)

    # align all chunks
    fields = da.map_overlap(
        align_single_block,
        fix_da, mov_da, fm_da, mm_da,
        depth=overlaps_list,
        dtype=object,
        boundary='none',
        trim=False,
        align_arrays=False,
        chunks=[1,1,1],
    ).compute()

    # reconstruct single transform
    transform = np.zeros(fix_p.shape + (3,), dtype=np.float32)

    # adds fields to transform
    block_grid = fields.shape[:3]
    for ijk in range(np.prod(block_grid)):
        i, j, k = np.unravel_index(ijk, block_grid)
        x, y, z = (max(0, x*y-z) for x, y, z in zip((i, j, k), blocksize, overlaps))
        xr, yr, zr = fields[i, j, k].shape[:3]
        transform[x:x+xr, y:y+yr, z:z+zr] += fields[i, j, k][...]

    # crop back to original shape and return
    x, y, z = fix.shape
    return transform[:x, :y, :z]


@cluster
def nested_distributed_piecewise_alignment_pipeline(
    fix,
    mov,
    fix_spacing,
    mov_spacing,
    block_schedule,
    parameter_schedule=None,
    initial_transform_list=None,
    fix_mask=None,
    mov_mask=None,
    intermediates_path=None,
    cluster=None,
    cluster_kwargs={},
    **kwargs,
):
    """
    Nested piecewise affine alignments.
    Two levels of nesting: outer levels and inner levels.
    Transforms are averaged over inner levels and composed
    across outer levels. See the `block_schedule` parameter
    for more details.

    This method is good at capturing large bends and twists that
    cannot be captured with global rigid and affine alignment.

    Parameters
    ----------
    fix : ndarray
        the fixed image

    mov : ndarray
        the moving image; if `initial_transform_list` is None then
        `fix.shape` must equal `mov.shape`

    fix_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the fixed image.
        Length must equal `fix.ndim`

    mov_spacing : 1d array
        The spacing in physical units (e.g. mm or um) between voxels
        of the moving image.
        Length must equal `mov.ndim`

    block_schedule : list of lists of tuples of ints.
        Block structure for outer and inner levels.
        Tuples must all be of length `fix.ndim`

        Example:
            [ [(2, 1, 1), (1, 2, 1),],
              [(3, 1, 1), (1, 1, 2),],
              [(4, 1, 1), (2, 2, 1), (2, 2, 2),], ]

            This block schedule specifies three outer levels:
            1) This outer level contains two inner levels:
                1.1) Piecewise rigid+affine with 2 blocks along first axis
                1.2) Piecewise rigid+affine with 2 blocks along second axis
            2) This outer level contains two inner levels:
                2.1) Piecewise rigid+affine with 3 blocks along first axis
                2.2) Piecewise rigid+affine with 2 blocks along third axis
            3) This outer level contains three inner levels:
                3.1) Piecewise rigid+affine with 4 blocks along first axis
                3.2) Piecewise rigid+affine with 4 blocks total: the first
                     and second axis are each cut into 2 blocks
                3.3) Piecewise rigid+affine with 8 blocks total: all axes
                     are cut into 2 blocks

            1.1 and 1.2 are computed (serially) then averaged. This result
            is stored. 2.1 and 2.2 are computed (serially) then averaged.
            This is then composed with the result from the first level.
            This process proceeds for as many levels that are specified.

            Each instance of a piecewise rigid+affine alignment is handled
            by `distributed_piecewise_affine_alignment` and is therefore
            parallelized over blocks on distributed hardware.

    parameter_schedule : list of type dict (default: None)
        Overrides the general parameter `distributed_piecewise_affine_align`
        parameter settings for individual instances. Length of the list
        (total number of dictionaries) must equal the total number of
        tuples in `block_schedule`.

    initial_transform_list : list of ndarrays (default: None)
        A list of transforms to apply to the moving image before running
        twist alignment. If `fix.shape` does not equal `mov.shape`
        then an `initial_transform_list` must be given.

    fix_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the fixed image

    mov_mask : binary ndarray (default: None)
        A mask limiting metric evaluation region of the moving image

    intermediates_path : string (default: None)
        Path to folder where intermediate results are written.
        The deform, transformed moving image, and transformed
        moving image mask (if given) are stored on disk as npy files.
    
    cluster_kwargs : dict (default: {})
        Arguments passed to ClusterWrap.cluster
        If working with an LSF cluster, this will be
        ClusterWrap.janelia_lsf_cluster. If on a workstation
        this will be ClusterWrap.local_cluster.
        This is how distribution parameters are specified.

    kwargs : any additional arguments
        Passed to `distributed_piecewise_affine_align`

    Returns
    -------
    field : ndarray
        Composition of all outer level transforms. A displacement vector
        field of the shape `fix.shape` + (3,) where the last dimension
        is the vector dimension.
    """

    # set working copies of moving data
    if initial_transform_list is not None:
        current_moving = apply_transform(
            fix, mov, fix_spacing, mov_spacing,
            transform_list=initial_transform_list,
        )
        current_moving_mask = None
        if mov_mask is not None:
            current_moving_mask = apply_transform(
                fix, mov_mask, fix_spacing, mov_spacing,
                transform_list=initial_transform_list,
            )
            current_moving_mask = (current_moving_mask > 0).astype(np.uint8)
    else:
        current_moving = np.copy(mov)
        current_moving_mask = None if mov_mask is None else np.copy(mov_mask)

    # initialize container and Loop over outer levels
    counter = 0  # count each call to distributed_piecewise_affine_align
    deform = np.zeros(fix.shape + (3,), dtype=np.float32)
    for outer_level, inner_list in enumerate(block_schedule):

        # initialize inner container and Loop over inner levels
        ddd = np.zeros_like(deform)
        for inner_level, nblocks in enumerate(inner_list):

            # determine parameter settings
            if parameter_schedule is not None:
                instance_kwargs = {**kwargs, **parameter_schedule[counter]}
            else:
                instance_kwargs = kwargs

            # align
            ddd += distributed_piecewise_alignment_pipeline(
                fix, current_moving,
                fix_spacing, fix_spacing,  # images should be on same grid
                nblocks=nblocks,
                fix_mask=fix_mask,
                mov_mask=current_moving_mask,
                cluster=cluster,
                cluster_kwargs=cluster_kwargs,
                **instance_kwargs,
            )

            # increment counter
            counter += 1

        # take mean
        ddd = ddd / len(inner_list)

        # if not first iteration, compose with existing deform
        if outer_level > 0:
            deform = compose_displacement_vector_fields(
                deform, ddd, fix_spacing,
            )
        else:
            deform = ddd

        # combine with initial transforms if given
        if initial_transform_list is not None:
            transform_list = initial_transform_list + [deform,]
        else:
            transform_list = [deform,]

        # update working copy of image
        current_moving = apply_transform(
            fix, mov, fix_spacing, mov_spacing,
            transform_list=transform_list,
        )
        # update working copy of mask
        if mov_mask is not None:
            current_moving_mask = apply_transform(
                fix, mov_mask, fix_spacing, mov_spacing,
                transform_list=transform_list,
            )
            current_moving_mask = (current_moving_mask > 0).astype(np.uint8)

        # write intermediates
        if intermediates_path is not None:
            ois = str(outer_level)
            deform_path = (intermediates_path + '/twist_deform_{}.npy').format(ois)
            image_path = (intermediates_path + '/twist_image_{}.npy').format(ois)
            mask_path = (intermediates_path + '/twist_mask_{}.npy').format(ois)
            np.save(deform_path, deform)
            np.save(image_path, current_moving)
            if mov_mask is not None:
                np.save(mask_path, current_moving_mask)

    # return deform
    return deform
    




import numpy as np
from scipy.ndimage.filters import gaussian_filter1d
from scipy.ndimage import rotate


def raw_moment(image, spacing, iord, jord, kord):
    """
    """

    nrows, ncols, nstacks = image.shape
    x, y, z = np.mgrid[:nrows, :ncols, :nstacks]
    x = x * spacing[0]
    y = y * spacing[1]
    z = z * spacing[2]
    return (image * x**iord * y**jord * z**kord).sum()


def principal_axes(image, spacing):
    """
    """

    # mean
    image_sum = image.sum()
    x_bar = raw_moment(image, spacing, 1, 0, 0)
    y_bar = raw_moment(image, spacing, 0, 1, 0)
    z_bar = raw_moment(image, spacing, 0, 0, 1)
    mean = np.array([x_bar, y_bar, z_bar]) / image_sum

    # covariance
    c00 = raw_moment(image, spacing, 2, 0, 0) - x_bar * mean[0]
    c11 = raw_moment(image, spacing, 0, 2, 0) - y_bar * mean[1]
    c22 = raw_moment(image, spacing, 0, 0, 2) - z_bar * mean[2]
    c01 = raw_moment(image, spacing, 1, 1, 0) - x_bar * mean[1]
    c02 = raw_moment(image, spacing, 1, 0, 1) - x_bar * mean[2]
    c12 = raw_moment(image, spacing, 0, 1, 1) - y_bar * mean[2]
    covariance = np.array([[c00, c01, c02], [c01, c11, c12],[c02, c12, c22]]) / image_sum

    # principal axes
    eigvals, eigvecs = np.linalg.eigh(covariance)

    # ensure positive axis orientation and consistent axis ordering
    for col in range(eigvecs.shape[-1]):
        idx = np.argmax(np.abs(eigvecs[:, col]))
        if eigvecs[idx, col] < 0:
            eigvecs[:, col] *= -1
    sort_idx = np.argmax(eigvecs, axis=1)
    return mean, eigvals[sort_idx], eigvecs[:, sort_idx]


def align_modes(mean1, cov1, mean2, cov2):
    """
    """

    # align centers of mass
    translation = np.eye(4)
    translation[:3, -1] = mean2 - mean1

    # rotate principal axes about center of mass
    com_right = np.eye(4)
    com_right[:3, -1] = -mean2
    com_left = np.abs(com_right)
    rot = np.eye(4)
    rot[:3, :3] = np.matmul(cov2, cov1.T)

    # compose transforms
    rotation = np.matmul(com_left, np.matmul(rot, com_right))
    return np.matmul(rotation, translation)


def sagittal_medial_polynomial(image, spacing, cor_axis, ax_axis, order=2):
    """
    """

    # get center of mass for each slice along axis
    orders = np.eye(3)
    points = np.empty((image.shape[cor_axis], 3))
    for idx in range(image.shape[cor_axis]):
        sli = [slice(None, None, None),]*3
        sli[cor_axis] = slice(idx, idx+1)
        image_slice = image[sli]
        norm = image_slice.sum()
        for ax in range(3):
            if ax != cor_axis:
                points[idx, ax] = raw_moment(image_slice, spacing, *orders[ax]) / norm
            else:
                points[idx, ax] = idx * spacing[cor_axis]

    # remove slices that had no data
    points = points[~np.isnan(points).any(axis=1)]

    # fit a smooth curve through those centers
    return np.polyfit(points[:, cor_axis], points[:, ax_axis], order)


def compute_arc_lengths(mask, smp, spacing, cor_axis, ax_axis, sag_plane):
    """
    """

    arc = 0
    result = []
    previous_slice = None
    ant_mark = 0
    post_mark = mask.shape[cor_axis] - 1
    for i in range(mask.shape[cor_axis]):
        # compute coordinates
        x = i * spacing[cor_axis]
        y = smp[0]*x**2 + smp[1]*x + smp[2]
        j = int(round( y / spacing[ax_axis] ))

        # accumulate arc length
        arc += (1 + (2*smp[0]*x + smp[1])**2)**0.5
        result.append([x, y, arc])

        # if entering or exiting foreground, mark index
        sli = [slice(sag_plane, sag_plane+1),]*3
        sli[cor_axis] = slice(i, i+1)
        sli[ax_axis] = slice(j, j+1)
        if previous_slice is not None:
            if mask[previous_slice] == 0 and mask[sli] == 1:
                ant_mark = i
            if mask[previous_slice] == 1 and mask[sli] == 0:
                post_mark = i-1
        previous_slice = sli

    # correspondence defined by anterior/posterior landmarks set
    # to 0/1 respectively
    result = np.array(result)
    result[:, 2] = result[:, 2] - result[ant_mark, 2]
    result[:, 2] = result[:, 2] / result[post_mark, 2]
    return result


def force_to_displacement(
    force,
    score_function,
    step=0.01,
    field_sigma=1.0,
):
    """
    """

    # integrate force with smoothing
    deform = np.zeros_like(force)
    score = score_function(deform)
    previous_score = score + 1
    while score - previous_score < 0:
        deform = deform + step*force
        for i in range(3):
            deform[..., i] = gaussian_filter1d(deform[..., i], field_sigma, axis=-1)
        previous_score = score
        score = score_function(deform)
    return deform


def align_sagittal_medial_polynomials(
    fixed, moving,
    fixed_smp, moving_smp,
    fixed_spacing, moving_spacing,
    cor_axis, ax_axis, sag_plane,
    force_sigma=64.0,
    **kwargs,
):
    """
    """

    # get coordinates and arc lengths
    fix_arc = compute_arc_lengths(
        fixed, fixed_smp, fixed_spacing,
        cor_axis, ax_axis, sag_plane,
    )
    mov_arc = compute_arc_lengths(
        moving, moving_smp, moving_spacing,
        cor_axis, ax_axis, sag_plane,
    )

    # establish correspondences
    delta_x = np.interp(fix_arc[:, -1], mov_arc[:, -1], mov_arc[:, 0]) - fix_arc[:, 0]
    delta_y = np.interp(fix_arc[:, -1], mov_arc[:, -1], mov_arc[:, 1]) - fix_arc[:, 1]
    delta_z = np.zeros_like(delta_x)
    delta = np.hstack((delta_z[..., None], delta_x[..., None], delta_y[..., None]))

    # get voxel coordinates of smp curves
    x_as_i = np.round(fix_arc[:, 0] / fixed_spacing[cor_axis]).astype(np.uint16)
    y_as_j = np.round(fix_arc[:, 1] / fixed_spacing[ax_axis]).astype(np.uint16)
    ant_mark = np.argwhere(fix_arc[:, -1] == 0.)[0, 0]
    post_mark = np.argwhere(fix_arc[:, -1] == 1.)[0, 0]
    x_as_i_fg = x_as_i[ant_mark:post_mark+1]  # within foreground
    y_as_j_fg = y_as_j[ant_mark:post_mark+1]

    # define force
    force = np.zeros(fixed.shape + (3,), dtype=np.float32)
    force[:, x_as_i, y_as_j] = delta
    for i in range(3):
        force[..., i] = gaussian_filter1d(force[..., i], force_sigma, axis=-1)
    scale = delta[x_as_i_fg, 1:] / force[sag_plane, x_as_i_fg, y_as_j_fg, 1:]
    force = force * np.mean(scale)

    # define integration convergence criteria
    def score_function(x):
        dist = x[sag_plane, x_as_i_fg, y_as_j_fg] - delta[x_as_i_fg]
        return np.max( np.linalg.norm(dist, axis=-1) )

    # integrate force to smooth displacement and return
    return force_to_displacement(force, score_function, **kwargs)





def symmetric_padding(image, center):
    """
    """

    pads = []
    for i in range(len(image.shape)):
        left = center[i]
        right = image.shape[i] - center[i] - 1
        if right - left > 0:
            pads.append((right-left, 0))
        elif left - right > 0:
            pads.append((0, left-right))
        else:
            pads.append((0, 0))
    return pads


def correlations(fix_window, mov_window):
    """
    """

    # ensure enough precision for sensitive calculations
    fix_window = fix_window.astype(np.float64)  # W
    mov_window = mov_window.astype(np.float64)  # NxW

    # get statistics
    fix_mean = np.mean(fix_window)  # scalar
    fix_std  = np.std(fix_window)  # scalar
    mov_mean = np.mean(mov_window, axis=1)  # N
    mov_std  = np.std(mov_window, axis=1)  # N

    # center arrays
    fix_center = fix_window - fix_mean  # W
    mov_center = mov_window - mov_mean[..., None]  # NxW

    # compute correlations
    correlations = np.sum(fix_center[None, ...] * mov_center, axis=1) / len(fix_window)  # N
    return correlations / mov_std / fix_std  # N


def brute_force_rotation_2d(
    fixed, moving, moving_mask,
    min_angle, max_angle, step,
    threshold=0.4,
):
    """
    """

    # only attempt to align planes with some coincident data
    if not np.any( fixed*moving ):
        return 0, np.zeros(2)

    # get moving center of mass in pixel units
    denom = np.sum(moving_mask)
    x_bar = raw_moment(moving_mask[..., None], np.ones(3), 1, 0, 0)
    y_bar = raw_moment(moving_mask[..., None], np.ones(3), 0, 1, 0)
    center_pix = np.round( np.array([x_bar, y_bar]) / denom ).astype(int)

    # pad moving image for symmetry about center
    pads = symmetric_padding(moving_mask, center_pix)
    sli = []
    for p in pads:
        if p[0] != 0:
            sli.append(slice(p[0], None))
        elif p[1] != 0:
            sli.append(slice(None, -p[1]))
        else:
            sli.append(slice(None, None))
    moving_pad = np.pad(moving, pads)

    # make all rotations
    fixed_flat = fixed.flatten()
    angles = np.arange(min_angle, max_angle+step/2., step)
    rotated = np.empty((len(angles), len(fixed_flat)))
    for i, angle in enumerate(angles):
        rotated[i] = rotate(moving_pad, angle, reshape=False, order=1)[sli].flatten()

    # find best correlation
    corrs = correlations(fixed_flat, rotated)
    best_idx = np.argmax(corrs)
    best_corr, best_angle = corrs[best_idx], angles[best_idx]
    if best_corr < threshold:
        best_corr, best_angle = 0, 0

    # return rigid transform
    return best_angle, center_pix


def rigid_matrix(angle, center):
    """
    """

    # construct rotation matrix
    rot = np.eye(3)
    rot[0, 0] =  np.cos(angle * np.pi/180)
    rot[1, 1] =  np.cos(angle * np.pi/180)
    rot[0, 1] =  np.sin(angle * np.pi/180)
    rot[1, 0] = -np.sin(angle * np.pi/180)

    # construct translations to/from center of rotation
    trans_right = np.eye(3)
    trans_right[:2, -1] = -center
    trans_left = np.abs(trans_right)

    return np.matmul(trans_left, np.matmul(rot, trans_right))


def align_twist(
    fixed, moving, moving_mask,
    spacing, axis,
    min_angle, max_angle, angle_step,
    smooth_sigma=12.0,
):
    """
    """

    # get rotation for each slice
    angles, centers = [], []
    for i in range(fixed.shape[axis]):
        slc = [slice(None, None),]*3
        slc[axis] = slice(i, i+1)
        f = fixed[slc].squeeze()
        m = moving[slc].squeeze()
        mm = moving_mask[slc].squeeze()
        angle, center = brute_force_rotation_2d(
            f, m, mm, min_angle, max_angle, angle_step,
        )
        angles.append(angle)
        centers.append(center)

    # smooth angles and centers
    angles, centers = np.array(angles), np.array(centers)
    angles = gaussian_filter1d(angles, smooth_sigma)
    centers = gaussian_filter1d(centers, smooth_sigma, axis=0)

    # convert to vector field
    deform = np.empty(fixed.shape + (3,))
    nrows, ncols, nstacks = fixed.shape
    grid = np.array(np.mgrid[:nrows, :stacks]).transpose(1,2,0)
    for i in range(deform.shape[axis]):

        # get slice
        slc = [slice(None, None),]*4
        slc[axis] = slice(i, i+1)
        if axis == 0:
            slc[-1] = slice(1, 3)
        elif axis == 1:
            slc[-1] = slice(0, 3, 2)
        elif axis == 2:
            slc[-1] = slice(0, 2)

        # transform grid
        rigid = rigid_matrix(angles[i], centers[i])
        mm, tt = rigid[:2, :2], rigid[:2, -1]
        res = (np.einsum('...ij,...j->...i', mm, grid) + tt) - grid
        deform[slc] = np.expand_dims(res, axis)

    return deform * spacing





import setuptools

setuptools.setup(
    name="CircuitSeeker",
    version="0.4.3",
    author="Greg M. Fleishman",
    author_email="greg.nli10me@gmail.com",
    description="Tools for finding neural circuits",
    url="https://github.com/GFleishman/CircuitSeeker",
    license="MIT",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy',
        'scipy',
        'h5py',
        'bokeh',
        'dask',
        'dask[array]',
        'dask[bag]',
        'dask[delayed]',
        'dask[distributed]',
        'dask-jobqueue',
        'SimpleITK',
        'zarr',
        'numcodecs',
        'morphsnakes',
        'ClusterWrap>=0.1.6',
        'pynrrd',
        'greedypy',
        'dask-stitch>=0.2.1',
    ]
)



import numpy as np
import SimpleITK as sitk
from CircuitSeeker._wrap_irm import configure_irm
import CircuitSeeker.utility as ut


def patch_mutual_information(
    fix,
    mov,
    spacing,
    radius,
    stride,
    percentile_cutoff=0,
    fix_mask=None,
    mov_mask=None,
    return_metric_image=False,
    **kwargs,
):
    """
    """

    # create sitk versions of data
    fix_sitk = ut.numpy_to_sitk(fix.transpose(2, 1, 0), spacing[::-1])
    fix_sitk = sitk.Cast(fix_sitk, sitk.sitkFloat32)
    mov_sitk = ut.numpy_to_sitk(mov.transpose(2, 1, 0), spacing[::-1])
    mov_sitk = sitk.Cast(mov_sitk, sitk.sitkFloat32)

    # convert to voxel units
    radius = np.round(radius / spacing).astype(np.uint16)
    stride = np.round(stride / spacing).astype(np.uint16)

    # determine patch sample centers
    samples = np.zeros_like(fix)
    sample_points = tuple(slice(r, -r, s) for r, s in zip(radius, stride))
    samples[sample_points] = 1

    # mask sample points
    if fix_mask is not None: samples = samples * fix_mask
    if mov_mask is not None: samples = samples * mov_mask

    # convert to list of coordinates
    samples = np.column_stack(np.nonzero(samples))

    # create irm for evaluating metric
    irm = configure_irm(**kwargs)

    # create container for metric image
    if return_metric_image:
        metric_image = np.zeros(fix.shape, dtype=np.float32)

    # loop over patches and evaluate
    scores = []
    for sample in samples:

        # get patches
        patch = tuple(slice(s-r, s+r+1) for s, r in zip(sample, radius))
        f = fix_sitk[patch]
        m = mov_sitk[patch]

        # evaluate metric
        try:
            scores.append( irm.MetricEvaluate(f, m) )
        except Exception as e:
            scores.append( 0 )

        # update metric image
        if return_metric_image:
            metric_image[patch] = scores[-1]

    # threshold scores
    scores = np.array(scores)
    if percentile_cutoff > 0:
        cutoff = np.percentile(-scores, percentile_cutoff)
        scores = scores[-scores > cutoff]

    # return results
    if return_metric_image:
        return np.mean(scores), metric_image
    else:
        return np.mean(scores)

